/***********************************************************************
 *
 * Copyright (C) 2009 Graeme Gott <graeme@gottcode.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ***********************************************************************/

#include "board.h"

#include "hole.h"
#include "movement.h"
#include "peg.h"

#include <QLinearGradient>
#include <QRegExp>
#include <QResizeEvent>
#include <QSettings>
#include <QUndoStack>

#include <algorithm>

/*****************************************************************************/

inline uint qHash(const QPoint& key) {
	return (key.x() << 16) + key.y();
}

/*****************************************************************************/

Board::Board(QUndoStack* moves, QWidget* parent)
: QGraphicsView(parent),
  m_status(0),
  m_moves(moves) {
	QGraphicsScene* scene = new QGraphicsScene(this);
	setScene(scene);

	// Configure view
	setCacheMode(QGraphicsView::CacheBackground);
	setRenderHints(QPainter::Antialiasing | QPainter::TextAntialiasing);
	setFrameStyle(QFrame::NoFrame);
	setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
	setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOff);

	// Start game
	QSettings settings;
	int seed = settings.value("Current/Seed", 0).toInt();
	int difficutly = settings.value("Current/Difficulty", settings.value("Difficulty", 1).toInt()).toInt();
	QStringList movements = settings.value("Current/Moves").toStringList();
	generate(seed, difficutly);

	// Load current game
	QRegExp parse("(\\d+)x(\\d+) to (\\d+)x(\\d+)");
	foreach (QString movement, movements) {
		if (!parse.exactMatch(movement)) {
			continue;
		}

		QPoint old_hole(parse.cap(1).toInt(), parse.cap(2).toInt());
		QPoint new_hole(parse.cap(3).toInt(), parse.cap(4).toInt());
		if (isPeg(old_hole) && isHole(new_hole)) {
			move(old_hole, new_hole);
		} else {
			qWarning("Invalid move: %dx%d to %dx%d", old_hole.x(), old_hole.y(), new_hole.x(), new_hole.y());
		}
	}
}

/*****************************************************************************/

bool Board::isHole(const QPoint& hole) const {
	return m_holes.contains(hole) ? !m_holes.value(hole)->hasPeg() : 0;
}

/*****************************************************************************/

bool Board::isPeg(const QPoint& hole) const {
	return m_holes.contains(hole) ? m_holes.value(hole)->hasPeg() : 0;
}

/*****************************************************************************/

Hole* Board::hole(const QPoint& hole) const {
	return m_holes.contains(hole) ? m_holes[hole] : 0;
}

/*****************************************************************************/

void Board::generate(int seed, int difficulty) {
	// Remove old board
	m_status = 0;
	m_moves->clear();
	m_holes.clear();
	scene()->clear();
	setInteractive(true);

	// Set seed
	if (seed == 0) {
		srand(time(0));
		seed = rand();
	}
	srand(seed);
	window()->setWindowFilePath(QString::number(seed));

	// Store game settings
	QSettings settings;
	settings.setValue("Current/Seed", seed);
	settings.setValue("Current/Difficulty", difficulty);
	settings.remove("Current/Moves");

	// Create board
	QHash<QPoint, int> board;

	QList<QPoint> directions = QList<QPoint>() << QPoint(-1,0) << QPoint(-1,-1) << QPoint(0,-1) << QPoint(1,-1) << QPoint(1,0) << QPoint(1,1) << QPoint(0,1) << QPoint(-1,1);
	QList<QPoint> holes;

	QPoint start(0, 0);
	board.insert(start, 2);
	holes.append(start);

	int x1 = 0;
	int y1 = 0;
	int x2 = 0;
	int y2 = 0;

	difficulty = qBound(1, difficulty, 101);
	difficulty += 5;
	for (int i = 0; i < difficulty; ++i) {
		for (int j = 0; j < holes.count(); ++j) {
			const QPoint& hole = holes.at(j);

			std::random_shuffle(directions.begin(), directions.end());
			for (int k = 0; k < directions.count(); ++k) {
				const QPoint& direction = directions.at(k);

				QPoint hole1 = direction + hole;
				QPoint hole2 = (direction * 2) + hole;
				if ( (board.value(hole1) != 2) && (board.value(hole2) != 2) ) {
					board[hole] = 1;
					board[hole1] = 2;
					board[hole2] = 2;
					holes.removeOne(hole);
					holes.append(hole1);
					holes.append(hole2);

					x1 = qMin(x1, hole2.x());
					x2 = qMax(x2, hole2.x());
					y1 = qMin(y1, hole2.y());
					y2 = qMax(y2, hole2.y());

					k = directions.count();
					j = holes.count();
				}
			}
		}

		std::random_shuffle(holes.begin(), holes.end());
	}

	// Create scene
	int width = (x2 - x1) + 1;
	int height = (y2 - y1) + 1;
	setSceneRect((x1 * 20) - 10, (y1 * 20) - 10, (width * 20) + 20, (height * 20) + 20);
	scaleHoles(size());

	QHashIterator<QPoint, int> i(board);
	while (i.hasNext()) {
		i.next();
		QPoint pos = i.key();
		Hole* hole = 0;

		if (i.value() > 0) {
			hole = new Hole(pos.x(), pos.y());
			scene()->addItem(hole);
		}
		if (i.value() == 2) {
			Peg* peg = new Peg(pos.x(), pos.y(), this);
			hole->setPeg(peg);
			scene()->addItem(peg);
		}

		if (hole != 0) {
			m_holes.insert(pos, hole);
		}
	}
}

/*****************************************************************************/

void Board::move(const QPoint& old_hole, const QPoint& new_hole) {
	// Move peg
	Movement* movement = new Movement(old_hole, new_hole, this);
	m_moves->push(movement);

	// Handle finishing the game
	if (checkFinished()) {
		setInteractive(false);
		scene()->update();
		m_moves->clear();
		QSettings().remove((m_status == 2) ? "Current" : "Current/Moves");
	}
}

/*****************************************************************************/

void Board::setHighlight(const QList<QPoint>& holes, bool highlight) {
	foreach (const QPoint& hole, holes) {
		Q_ASSERT(m_holes.contains(hole));
		m_holes[hole]->setHighlight(highlight);
	}
}

/*****************************************************************************/

void Board::drawBackground(QPainter* painter, const QRectF& rect) {
	QLinearGradient gradient(rect.topLeft(), rect.bottomRight());
	gradient.setColorAt(0, QColor(241, 208, 178));
	gradient.setColorAt(1, QColor(193, 146, 70));
	painter->fillRect(rect, gradient);
}

/*****************************************************************************/

void Board::drawForeground(QPainter* painter, const QRectF&) {
	if (m_status) {
		QString message((m_status == 2) ? tr("Success") : tr("Game Over"));
		QFontMetrics metrics(QFont("Sans", 24));
		int w = metrics.width(message);
		int h = metrics.height();
		QPixmap pixmap(QSize(w + h, h * 2));
		pixmap.fill(QColor(0, 0, 0, 0));
		{
			QPainter pixmap_painter(&pixmap);

			pixmap_painter.setPen(Qt::NoPen);
			pixmap_painter.setBrush(QColor(0, 0, 0, 200));
			pixmap_painter.setRenderHint(QPainter::Antialiasing, true);
			pixmap_painter.drawRoundedRect(0, 0, w + h, h * 2, 10, 10);

			pixmap_painter.setFont(QFont("Sans", 24));
			pixmap_painter.setPen(Qt::white);
			pixmap_painter.setRenderHint(QPainter::TextAntialiasing, true);
			pixmap_painter.drawText(h / 2, h / 2 + metrics.ascent(), message);
		}

		painter->save();
		painter->resetTransform();
		painter->drawPixmap((width() - pixmap.width()) / 2, (height() - pixmap.height()) / 2, pixmap);
		painter->restore();
	}
}

/*****************************************************************************/

void Board::resizeEvent(QResizeEvent* event) {
	scaleHoles(event->size());
	QGraphicsView::resizeEvent(event);
}

/*****************************************************************************/

void Board::scaleHoles(const QSize& board_size) {
	QSizeF size = sceneRect().size();
	size.scale(board_size, Qt::KeepAspectRatio);
	float scale_factor = qMax(size.width() / sceneRect().width(), size.height() / sceneRect().height());

	resetTransform();
	scale(scale_factor, scale_factor);
}

/*****************************************************************************/

bool Board::checkFinished() {
	int pegs = 0;
	foreach (Hole* hole, m_holes) {
		if (hole->hasPeg()) {
			pegs++;
			if (hole->peg()->canMove()) {
				return false;
			}
		}
	}
	m_status = (pegs == 1) ? 2 : 1;
	return true;
}

/*****************************************************************************/
