/***********************************************************************
 *
 * Copyright (C) 2009 Graeme Gott <graeme@gottcode.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ***********************************************************************/

#include "window.h"

#include "board.h"

#include <QApplication>
#include <QDialog>
#include <QDialogButtonBox>
#include <QFormLayout>
#include <QMenuBar>
#include <QMessageBox>
#include <QSettings>
#include <QSpinBox>
#include <QUndoStack>
#include <QVBoxLayout>

/*****************************************************************************/

Window::Window()
: m_seed(0),
  m_difficulty(QSettings().value("Difficulty", 5).toInt()) {
#if defined(Q_OS_UNIX) && !defined(Q_OS_MAC)
	setWindowIcon(QIcon(":/peg-e.png"));
#endif

	// Create board
	QUndoStack* moves = new QUndoStack(this);
	m_board = new Board(moves, this);
	m_seed = QSettings().value("Current/Seed").toInt();
	setCentralWidget(m_board);

	// Create menus
	QMenu* menu = menuBar()->addMenu(tr("&Game"));
	menu->addAction(tr("&New"), this, SLOT(newGame()), tr("Ctrl+N"));
	menu->addAction(tr("&Restart"), this, SLOT(restartGame()), tr("F5"));
	menu->addSeparator();
	QAction* action = moves->createUndoAction(menu);
	action->setShortcut(tr("Ctrl+Z"));
	menu->addAction(action);
	action = moves->createRedoAction(menu);
	action->setShortcut(tr("Shift+Ctrl+Z"));
	menu->addAction(action);
	menu->addSeparator();
	menu->addAction(tr("&Quit"), qApp, SLOT(quit()), tr("Ctrl+Q"));

	menu = menuBar()->addMenu(tr("&Help"));
	menu->addAction(tr("&About"), this, SLOT(about()));
	menu->addAction(tr("About &Qt"), qApp, SLOT(aboutQt()));

	// Restore size and position
	setMinimumSize(250, 250);
	resize(400, 400);
	restoreGeometry(QSettings().value("Geometry").toByteArray());
}

/*****************************************************************************/

void Window::closeEvent(QCloseEvent* event) {
	QSettings().setValue("Geometry", saveGeometry());
	QMainWindow::closeEvent(event);
}

/*****************************************************************************/

void Window::newGame() {
	QDialog dialog(this);
	dialog.setWindowTitle(tr("New Game"));

	// Create spinboxes
	QSpinBox* seed_box = new QSpinBox(&dialog);
	seed_box->setRange(0, RAND_MAX);
	seed_box->setValue(0);
	seed_box->setSpecialValueText(tr("Random"));

	QSpinBox* difficulty_box = new QSpinBox(&dialog);
	difficulty_box->setRange(1, 100);
	difficulty_box->setValue(m_difficulty);

	// Layout spinboxes
	QFormLayout* box_layout = new QFormLayout;
	box_layout->addRow(tr("Seed:"), seed_box);
	box_layout->addRow(tr("Difficulty:"), difficulty_box);

	// Create dialog buttons
	QDialogButtonBox* buttons = new QDialogButtonBox(QDialogButtonBox::Ok | QDialogButtonBox::Cancel, Qt::Horizontal, &dialog);
	connect(buttons, SIGNAL(accepted()), &dialog, SLOT(accept()));
	connect(buttons, SIGNAL(rejected()), &dialog, SLOT(reject()));

	// Layout dialog
	QVBoxLayout* layout = new QVBoxLayout(&dialog);
	layout->addLayout(box_layout);
	layout->addWidget(buttons);

	if (dialog.exec() == QDialog::Accepted) {
		// Start a new game
		m_board->generate(seed_box->value(), difficulty_box->value());

		// Track seed and difficulty values
		QSettings settings;
		m_seed = settings.value("Current/Seed").toInt();
		m_difficulty = settings.value("Current/Difficulty").toInt();
		settings.setValue("Difficulty", m_difficulty);
	}
}

/*****************************************************************************/

void Window::restartGame() {
	if (m_board->isFinished() || QMessageBox::question(this, tr("Question"), tr("Do you want to restart?"), QMessageBox::Yes | QMessageBox::No, QMessageBox::Yes) == QMessageBox::Yes) {
		m_board->generate(m_seed, m_difficulty);
	}
}

/*****************************************************************************/

void Window::about() {
	QMessageBox::about(this, tr("About Peg-E"), tr("<center><big><b>Peg-E 1.0.0</b></big><br/>Peg elimination game<br/><small>Copyright &copy; 2009 Graeme Gott</small></center>"));
}

/*****************************************************************************/
